from oe_databank.async_client import DatabankAsyncClient
from oe_databank.client import DatabankClient
from oe_databank.enums import (
    AvalancheSelectionFormat,
    FileFormat,
    Frequency,
    ListingType,
    Order,
    SelectionSortOrder,
    SelectionType,
    Sequence,
)
from oe_databank.models import (
    ColumnMetadata,
    DatabankListResponse,
    DatabankMetadata,
    DownloadRequestRegion,
    DownloadRequestVariable,
    FileDownloadRequestDto,
    QueueDownloadResponse,
    RegionMetadata,
    RegionResponse,
    Selection,
    TreeMetadata,
    TreeNode,
    TreeResponse,
)



"""
API data models for the databank. See more at https://services.oxfordeconomics.com/swagger#/.

"""

import datetime

from pydantic import BaseModel, Field, RootModel

from oe_databank.enums import (
    FileFormat,
    Frequency,
    ListingType,
    Order,
    SelectionSortOrder,
    SelectionType,
    Sequence,
)


class DownloadRequestRegion(BaseModel):
    """A region definition passed in a download request."""

    databankCode: str | None
    regionCode: str | None


class DownloadRequestVariable(BaseModel):
    """A variable definition passed in a download request."""

    variableCode: str | None
    productTypeCode: str | None
    measureCodes: list[str]


class Selection(BaseModel):
    """A databank selection. Nullable columns default to None."""

    downloadUrl: str | None = Field(default=None, exclude=True)
    id: str | None = Field(default=None, exclude=True)
    selectionType: SelectionType
    lastUpdate: datetime.datetime | None = Field(default=None, exclude=True)
    isTemporarySelection: bool
    databankCode: str | None = None
    sequence: Sequence
    groupingMode: bool
    transposeColumns: bool | None = None
    order: Order
    indicatorSortOrder: SelectionSortOrder
    locationSortOrder: SelectionSortOrder
    sortedColumnName: str | None = None
    sortedColumnOrder: str | None = None
    format: int
    legacyDatafeedFileStructure: bool
    variables: list[DownloadRequestVariable] = Field(default_factory=list)
    regions: list[DownloadRequestRegion] = Field(default_factory=list)
    contactId: str | None = Field(default=None, exclude=True)
    listingType: ListingType
    isDataFeed: bool
    name: str | None = None
    startYear: int
    endYear: int
    precision: int
    frequency: Frequency
    stackedQuarters: bool


class FileDownloadRequestDto(BaseModel):
    """A request to query multiple selections and download the results as a file."""

    selections: list[Selection] = Field(min_length=1)
    format: FileFormat
    name: str | None = None


class QueueDownloadResponse(BaseModel):
    """A response from the /QueueDownload endpoint."""

    Url: str
    ReadyUrl: str


class TreeMetadata(BaseModel):
    """The metadata for a tree in the databank."""

    Name: str
    TreeCode: str
    Url: str


class ColumnMetadata(BaseModel):
    """The metadata for a column in the databank."""
    ColumnName: str
    DisplayOrder: int
    MetadataFieldName: str


class DatabankMetadata(BaseModel):
    """The metadata for a databank."""

    Name: str
    Url: str
    DatabankCode: str
    StartYear: int
    EndYear: int
    HasQuarterlyData: bool
    Trees: list[TreeMetadata]
    MapUrl: str
    DatabankColumns: list[ColumnMetadata]
    HasAccess: bool


class DatabankListResponse(RootModel):
    """The response from the /Databank endpoint."""

    root: list[DatabankMetadata]

    def __iter__(self):
        return iter(self.root)

    def __getitem__(self, item):
        return self.root[item]

    def __len__(self):
        return len(self.root)


class TreeNode(BaseModel):
    """A node in the databank tree."""

    Name: str
    Code: str
    ProductTypeCode: str
    DisplayOrder: int
    HasAccess: bool
    Children: list["TreeNode"] = Field(default_factory=list)


class TreeResponse(RootModel):
    """The response from the /DatabankTree endpoint."""

    root: list[TreeNode]

    def __iter__(self):
        return iter(self.root)

    def __getitem__(self, item):
        return self.root[item]

    def __len__(self):
        return len(self.root)


class RegionMetadata(BaseModel):
    """The metadata for a region in the databank."""

    RegionCode: str
    Name: str
    DatabankCode: str
    DisplayOrder: int
    HasAccess: bool


class RegionResponse(BaseModel):
    """The response from the /Region endpoint."""

    Regions: list[RegionMetadata]



"""
An asynchronous API for interacting with the OE databank.
"""

import asyncio
import os
import time
import warnings

import httpx
import httpx._config
import httpx._types
import orjson

from oe_databank.models import DatabankListResponse, FileDownloadRequestDto, QueueDownloadResponse, RegionResponse, TreeResponse
from oe_databank.utils import download_response_to_path_async

# Request timeout for downloads
DEFAULT_DOWNLOAD_TIMEOUT_SECONDS = 60
# Total time to wait for a download to be ready
DEFAULT_POLL_TIMEOUT_SECONDS = 120
# Default polling interval
DEFAULT_POLL_INTERVAL_SECONDS = 5.0


class DatabankAsyncClient:
    """An asynchronous client for interacting with the Oxford Economics databank API.

    ```python
    import asyncio
    from oe_databank import (
        DatabankAsyncClient,
        FileDownloadRequestDto,
        FileFormat,
        Frequency,
        ListingType,
        Order,
        Selection,
        SelectionSortOrder,
        SelectionType,
        Sequence,
    )

    client = DatabankAsyncClient(api_key=...)

    request = FileDownloadRequestDto(
        selections=[
            Selection(
                selectionType=SelectionType.QUERY,
                isTemporarySelection=True,
                databankCode="IST",
                sequence=Sequence.EARLIEST_TO_LATEST,
                groupingMode=False,
                transposeColumns=False,
                order=Order.LOCATION_INDICATOR,
                indicatorSortOrder=SelectionSortOrder.ALPHABETICAL,
                locationSortOrder=SelectionSortOrder.ALPHABETICAL,
                format=0,
                legacyDatafeedFileStructure=False,
                variables=[],
                regions=[],
                listingType=ListingType.SHARED,
                isDataFeed=False,
                startYear=2023,
                endYear=2027,
                precision=5,
                frequency=Frequency.ANNUAL,
                stackedQuarters=False,
            )
        ],
        format=FileFormat.CSV,
        name="test-te-pull",
    )

    asyncio.run(client.query(request=request, to_path="result.csv"))
    ```

    """

    def __init__(
        self,
        client: httpx.AsyncClient | None = None,
        api_key: str | None = None,
        headers: dict | None = None,
        timeout: httpx._types.TimeoutTypes = httpx._config.DEFAULT_TIMEOUT_CONFIG,
        download_timeout_seconds: float | int = DEFAULT_DOWNLOAD_TIMEOUT_SECONDS,
        poll_timeout_seconds: float | int = DEFAULT_POLL_TIMEOUT_SECONDS,
        poll_interval_seconds: float | int = DEFAULT_POLL_INTERVAL_SECONDS,
    ):
        """Initialize the API.

        Args:
            client (httpx.Client, optional): An HTTP client. Defaults to None. If not provided, a client will be created with the provided API key.
            api_key (str, optional): The API key. Defaults to None. Must be provided here or as a header if a client is not provided.
            headers (dict, optional): Additional headers. Defaults to None. If an API key is provided here, it will override the `api_key` attribute.
            timeout (TimeoutTypes, optional): The non-download request timeout. Defaults to httpx._config.DEFAULT_TIMEOUT_CONFIG.
            download_timeout_seconds (float | int, optional): The total time to wait for a download. Defaults to the client's default.
            poll_timeout_seconds (float | int, optional): The total time to wait for a file to be ready. Defaults to the client's default.
            poll_interval_seconds (float | int, optional): The polling interval in seconds. Defaults to the client's default.
        """
        self.api_key = api_key
        self.headers = headers or {}
        self._base_url = "https://services.oxfordeconomics.com/api"
        # Define before setting the client
        self.timeout = timeout
        self._client = client or self._make_client()
        self.default_download_timeout_seconds = download_timeout_seconds
        self.default_poll_timeout_seconds = poll_timeout_seconds
        self.default_poll_interval_seconds = poll_interval_seconds

    @property
    def raw(self) -> httpx.AsyncClient:
        """The underlying HTTP client."""
        return self._client

    def _make_client(self) -> httpx.AsyncClient:
        """Create an HTTP client with the API key in the headers."""
        headers = self.headers.copy()
        if "api-key" in headers:
            if self.api_key is not None:
                warnings.warn(
                    "The 'api-key' header will override the 'api_key' attribute."
                )
        else:
            assert (
                self.api_key is not None
            ), "The API key must be provided as a header or attribute."
            headers["api-key"] = self.api_key
        return httpx.AsyncClient(
            base_url=self._base_url, headers=headers, timeout=self.timeout
        )

    def _resolve_download_timeout(self, timeout: float | int | None) -> float | int:
        """Resolve the download timeout."""
        if timeout is None:
            return self.default_download_timeout_seconds
        return timeout

    def _resolve_poll_timeout(self, timeout: float | int | None) -> float | int:
        """Resolve the poll timeout."""
        if timeout is None:
            return self.default_poll_timeout_seconds
        return timeout

    def _resolve_poll_interval(self, interval: float | int | None) -> float | int:
        """Resolve the poll interval."""
        if interval is None:
            return self.default_poll_interval_seconds
        return interval

    async def download_file_with_request_model(
        self,
        request: FileDownloadRequestDto,
        timeout: int | None = httpx.USE_CLIENT_DEFAULT,
        to_path: os.PathLike | None = None,
    ) -> bytes | None:
        """Download a file with a file download request in the body."""
        # Must follow redirects as the polling URL is returned until the download is ready.
        r = await self._client.post(
            "/filedownload",
            content=orjson.dumps(request.model_dump(mode="json")),
            timeout=self._resolve_download_timeout(timeout),
            follow_redirects=True,
        )
        r.raise_for_status()
        if to_path:
            return await download_response_to_path_async(r, to_path)
        return r.content

    async def download_file_with_selection_id(
        self,
        selection_id: str,
        timeout: int | None = httpx.USE_CLIENT_DEFAULT,
        to_path: os.PathLike | None = None,
    ) -> bytes | None:
        """Download a file by selection ID."""
        path = f"/filedownload/{selection_id}"
        r = await self._client.get(
            path, timeout=self._resolve_download_timeout(timeout), follow_redirects=True
        )
        r.raise_for_status()
        if to_path:
            return await download_response_to_path_async(r, to_path)
        return r.content

    async def download_file(
        self,
        *,
        selection_id: str | None = None,
        request: FileDownloadRequestDto | None = None,
        timeout: int | None = httpx.USE_CLIENT_DEFAULT,
        to_path: os.PathLike | None = None,
    ) -> bytes | None:
        """Download a file."""
        assert (
            selection_id or request
        ), "Either 'selection_id' or 'request' must be provided."
        if request:
            return await self.download_file_with_request_model(
                request=request, timeout=timeout, to_path=to_path
            )
        else:
            return await self.download_file_with_selection_id(
                selection_id=selection_id, timeout=timeout, to_path=to_path
            )

    async def queue_download_with_request_model(
        self,
        request: FileDownloadRequestDto,
        timeout: int | None = httpx.USE_CLIENT_DEFAULT,
    ) -> QueueDownloadResponse:
        """Queue a download with a file download request in the body."""
        r = await self._client.post(
            "/QueueDownload",
            content=orjson.dumps(request.model_dump(mode="json")),
            timeout=timeout,
        )
        r.raise_for_status()
        return QueueDownloadResponse.model_validate(orjson.loads(r.content))

    async def queue_download_with_selection_id(
        self,
        selection_id: str,
        filename: str | None = None,
        timeout: int | None = httpx.USE_CLIENT_DEFAULT,
    ) -> QueueDownloadResponse:
        """Queue a download request by selection ID."""
        path = f"/QueueDownload/{selection_id}" + (f"/{filename}" if filename else "")
        r = await self._client.get(path, timeout=timeout)
        r.raise_for_status()
        return QueueDownloadResponse.model_validate(orjson.loads(r.content))

    async def queue_download(
        self,
        *,
        selection_id: str | None = None,
        filename: str | None = None,
        request: FileDownloadRequestDto | None = None,
        timeout: int | None = httpx.USE_CLIENT_DEFAULT,
    ) -> QueueDownloadResponse:
        """Queue a download request."""
        assert (
            selection_id or request
        ), "Either 'selection_id' or 'request' must be provided."
        if request:
            return await self.queue_download_with_request_model(request, timeout)
        else:
            return await self.queue_download_with_selection_id(
                selection_id, filename, timeout
            )

    async def check_download_ready_with_id(
        self,
        download_id: str,
        timeout: int | None = httpx.USE_CLIENT_DEFAULT,
    ) -> bool:
        """Check if a download is ready.

        Args:
            download_id (str): The download ID.
            timeout (int, optional): The timeout. Defaults to httpx.USE_CLIENT_DEFAULT.

        Returns:
            bool: Whether the download is ready.
        """
        r = await self._client.get(f"/DownloadReady/{download_id}", timeout=timeout)
        r.raise_for_status()
        return orjson.loads(r.content)

    async def check_download_ready_with_queue_response(
        self,
        queue_response: QueueDownloadResponse,
        timeout: int | None = httpx.USE_CLIENT_DEFAULT,
    ) -> bool:
        """Check if a download is ready with a queue response.

        Args:
            queue_response (QueueDownloadResponse): The queue response.
            timeout (int, optional): The timeout. Defaults to httpx.USE_CLIENT_DEFAULT.

        Returns:
            bool: Whether the download is ready.
        """
        r = await self._client.get(queue_response.ReadyUrl, timeout=timeout)
        r.raise_for_status()
        return orjson.loads(r.content)

    async def download_generated_file_with_queue_response(
        self,
        queue_response: QueueDownloadResponse,
        timeout: int | None = httpx.USE_CLIENT_DEFAULT,
        to_path: os.PathLike | None = None,
    ) -> bytes | None:
        """Download a generated file with a queue response.

        Args:
            queue_response (QueueDownloadResponse): The queue response.
            timeout (int, optional): The timeout. Defaults to httpx.USE_CLIENT_DEFAULT.
            to_path (os.PathLike, optional): The path to save the file. Defaults to None.

        Returns:
            bytes | None: The file content if `to_path` is not provided.
        """
        r = await self._client.get(
            queue_response.Url,
            timeout=self._resolve_download_timeout(timeout),
            follow_redirects=True,
        )
        r.raise_for_status()
        if to_path:
            return await download_response_to_path_async(r, to_path)
        return r.content

    async def download_generated_file_with_id(
        self,
        download_id: str,
        timeout: int | None = httpx.USE_CLIENT_DEFAULT,
        to_path: os.PathLike | None = None,
    ) -> bytes | None:
        """Download a generated file by download ID.

        Args:
            download_id (str): The download ID.
            timeout (int, optional): The timeout. Defaults to httpx.USE_CLIENT_DEFAULT.
            to_path (os.PathLike, optional): The path to save the file. Defaults to None.
        """
        r = await self._client.get(
            f"/GenerateLink/{download_id}",
            timeout=self._resolve_download_timeout(timeout),
        )
        r.raise_for_status()
        if to_path:
            return await download_response_to_path_async(r, to_path)
        return r.content

    async def download_generated_file(
        self,
        *,
        download_id: str | None = None,
        queue_response: QueueDownloadResponse | None = None,
        timeout: int | None = httpx.USE_CLIENT_DEFAULT,
        to_path: os.PathLike | None = None,
    ) -> bytes | None:
        """Download a generated file.

        Args:
            download_id (str, optional): The download ID. Defaults to None.
            queue_response (QueueDownloadResponse, optional): The queue response. Defaults to None.
            timeout (int, optional): The timeout. Defaults to httpx.USE_CLIENT_DEFAULT.
            to_path (os.PathLike, optional): The path to save the file. Defaults to None.

        Returns:
            bytes | None: The file content if `to_path` is not provided.
        """
        assert (
            download_id or queue_response
        ), "Either 'download_id' or 'queue_response' must be provided."
        if queue_response:
            return await self.download_generated_file_with_queue_response(
                queue_response, timeout=timeout, to_path=to_path
            )
        else:
            return await self.download_generated_file_with_id(
                download_id, timeout=timeout, to_path=to_path
            )

    async def _poll(
        self,
        queue_response: QueueDownloadResponse,
        timeout: int | None = httpx.USE_CLIENT_DEFAULT,
        poll_timeout_seconds: float | int | None = None,
        poll_interval_seconds: float | int | None = None,
    ):
        """Poll a download until it is ready.

        Args:
            queue_response (QueueDownloadResponse): The queue response.
            timeout (int, optional): The request timeout. Defaults to httpx.USE_CLIENT_DEFAULT.
            poll_timeout_seconds (float | int | None, optional): The total time to wait for the download to be ready. Defaults to the client's default.
            poll_interval_seconds (float | int | None, optional): The polling interval in seconds. Defaults to the client's default.

        Raises:
            TimeoutError: If the download is not ready within the timeout.
        """
        poll_timeout_seconds = self._resolve_poll_timeout(poll_timeout_seconds)
        poll_interval_seconds = self._resolve_poll_interval(poll_interval_seconds)

        start_time = time.time()
        while (time.time() - start_time) < poll_timeout_seconds:
            if await self.check_download_ready_with_queue_response(
                queue_response, timeout
            ):
                return
            await asyncio.sleep(
                min(
                    poll_interval_seconds,
                    poll_timeout_seconds - (time.time() - start_time),
                    0.0,
                )
            )
        # Check one last time before returning
        if not await self.check_download_ready_with_queue_response(
            queue_response, timeout
        ):
            raise TimeoutError(
                f"Download not ready after {poll_timeout_seconds} seconds.",
                queue_response,
            )

    async def query(
        self,
        *,
        selection_id: str | None = None,
        request: FileDownloadRequestDto | None = None,
        to_path: os.PathLike | None = None,
        timeout: int | None = httpx.USE_CLIENT_DEFAULT,
        download_timeout_seconds: float | int | None = None,
        poll_timeout_seconds: float | int | None = None,
        poll_interval_seconds: float | int | None = None,
    ) -> bytes | None:
        """Query the databank for a selection. Must provide either `selection_id` or `request`. If `to_path` is provided, the file will be saved there and `None` will be returned. Otherwise, the file content will be returned as bytes.

        Args:
            selection_id (str, optional): The selection ID. Defaults to None.
            request (FileDownloadRequestDto, optional): The file download request. Defaults to None.
            to_path (os.PathLike, optional): The path to save the file. Defaults to None.
            timeout (int, optional): The request timeout. Defaults to httpx.USE_CLIENT_DEFAULT.
            download_timeout_seconds (float | int | None, optional): The total time to wait for the download to be ready. Defaults to the client's default.
            poll_timeout_seconds (float | int | None, optional): The total time to wait for the download to be ready. Defaults to the client's default.
            poll_interval_seconds (float | int | None, optional): The polling interval in seconds. Defaults to the client's default.

        Returns:
            bytes | None: The file content if `to_path` is not provided.

        Raises:
            TimeoutError: If the download is not ready within the download timeout.
        """
        if request:
            queue_response = await self.queue_download_with_request_model(
                request, timeout=timeout
            )
        else:
            queue_response = await self.queue_download_with_selection_id(
                selection_id, timeout=timeout
            )
        await self._poll(
            queue_response, timeout, poll_timeout_seconds, poll_interval_seconds
        )
        return await self.download_generated_file(
            queue_response=queue_response,
            timeout=download_timeout_seconds,
            to_path=to_path,
        )

    async def list_databanks(self, to_path: os.PathLike | None = None, timeout: int | None = None, as_json: bool = False) -> DatabankListResponse | list[dict] | None:
        """Get a list of all Oxford Economic databanks.

        Args:
            to_path (os.PathLike, optional): The path to save the file. Defaults to None.
            timeout (int, optional): The request timeout. Defaults to the client's default download timeout.
            as_json (bool, optional): Whether to return the response as native JSON types. Defaults to False.

        Returns:
            DatabankListResponse | list[dict] | None: The databank list if `to_path` is not provided. If `as_json` is `True`, the response will be a list of dictionaries.
        """
        r = await self._client.get("/Databank", timeout=timeout or self.default_download_timeout_seconds)
        r.raise_for_status()
        if to_path:
            return await download_response_to_path_async(r, to_path)
        if as_json:
            return orjson.loads(r.content)
        return DatabankListResponse.model_validate(orjson.loads(r.content))

    async def get_tree(self, tree_code: str, to_path: os.PathLike | None = None, timeout: int | None = None, as_json: bool = False) -> TreeResponse | list[dict] | None:
        """Get the tree metadata for a databank.

        Args:
            tree_code (str): The tree code.
            to_path (os.PathLike, optional): The path to save the file. Defaults to None.
            timeout (int, optional): The request timeout. Defaults to the client's default download timeout.
            as_json (bool, optional): Whether to return the response as native JSON types. Defaults to False.

        Returns:
            DatabankTreeResponse | list[dict] | None: The tree metadata if `to_path` is not provided. If `as_json` is `True`, the response will be a list of dictionaries.
        """
        r = await self._client.get(f"/tree/{tree_code}", timeout=timeout or self.default_download_timeout_seconds)
        r.raise_for_status()
        if to_path:
            return await download_response_to_path_async(r, to_path)
        if as_json:
            return orjson.loads(r.content)
        return TreeResponse.model_validate(orjson.loads(r.content))

    async def get_regions(self, databank_code: str, to_path: os.PathLike | None = None, timeout: int | None = None, as_json: bool = False) -> RegionResponse | dict | None:
        """Get the regions of a databank.

        Args:
            databank_code (str): The databank code.
            to_path (os.PathLike, optional): The path to save the file. Defaults to None.
            timeout (int, optional): The request timeout. Defaults to the client's default download timeout.
            as_json (bool, optional): Whether to return the response as native JSON types. Defaults to False.

        Returns:
            RegionResponse | dict | None: The region metadata if `to_path` is not provided. If `as_json` is `True`, the response will be a dictionary.
        """
        r = await self._client.get(f"/Region/{databank_code}", timeout=timeout or self.default_download_timeout_seconds)
        r.raise_for_status()
        if to_path:
            return await download_response_to_path_async(r, to_path)
        if as_json:
            return orjson.loads(r.content)
        return RegionResponse.model_validate(orjson.loads(r.content))



"""
Enumerations for the OE API. See more at https://services.oxfordeconomics.com/swagger#/.
"""

import enum


class AvalancheSelectionFormat(enum.IntEnum):
    """
    Enum for AvalancheSelectionFormat. Genuinely no idea what each means.

    Members:
        ZERO
        ONE
        TWO
        THREE
        FOUR
        FIVE
        SIX

    """

    ZERO = 0
    ONE = 1
    TWO = 2
    THREE = 3
    FOUR = 4
    FIVE = 5
    SIX = 6


class ListingType(enum.StrEnum):
    """
    Enum for ListingType.

    Members:
        COMPANY: Company.
        HIDDEN: Hidden.
        PRIVATE: Private.
        PUBLIC: Public.
        SHARED: Shared.
    """

    COMPANY = "Company"
    HIDDEN = "Hidden"
    PRIVATE = "Private"
    PUBLIC = "Public"
    SHARED = "Shared"


class FileFormat(enum.IntEnum):
    """
    Enum for FileFormat.

    Members:
        UNKNOWN: Unknown format.
        CSV: Comma-separated values. First line is "sep=,", header row second line, then data after.
    """

    UNKNOWN = 0
    CSV = 1


class Frequency(enum.StrEnum):
    """
    Enum for Frequency.

    Members:
        ANNUAL: Annual.
        MONTHLY: Monthly.
        QUARTERLY: Quarterly.
        BOTH: Both. Tests indicate 'Annual' and 'Quarterly'.
    """

    ANNUAL = "Annual"
    MONTHLY = "Monthly"
    QUARTERLY = "Quarterly"
    BOTH = "Both"


class Order(enum.StrEnum):
    """
    Enum for Order.

    Members:
        INDICATOR_LOCATION: Indicator then location.
        LOCATION_INDICATOR: Location then indicator.
    """

    INDICATOR_LOCATION = "IndicatorLocation"
    LOCATION_INDICATOR = "LocationIndicator"


class SelectionSortOrder(enum.StrEnum):
    """
    Enum for SelectionSortOrder.

    Members:
        ALPHABETICAL: Alphabetical order.
        TREE: Tree order.
    """

    ALPHABETICAL = "AlphabeticalOrder"
    TREE = "TreeOrder"


class SelectionType(enum.StrEnum):
    """
    Enum for SelectionType.

    Members:
        DATASERIES: Data series selection.
        QUERY: Query selection.
    """

    DATASERIES = "DataSeriesSelection"
    QUERY = "QuerySelection"


class Sequence(enum.StrEnum):
    """
    Enum for Sequence.

    Members:
        EARLIEST_TO_LATEST: Earliest to latest.
        LATEST_TO_EARLIEST: Latest to earliest.
    """

    EARLIEST_TO_LATEST = "EarliestToLatest"
    LATEST_TO_EARLIEST = "LatestToEarliest"



"""
Utility functions for the API.
"""

import os

import anyio
import httpx


def download_response_to_path(r: httpx.Response, path: os.PathLike):
    """Write a download response to a file."""
    with open(path, "wb") as f:
        for chunk in r.iter_bytes():
            f.write(chunk)


async def download_response_to_path_async(r: httpx.Response, path: os.PathLike):
    """Write a download response to a file."""
    async with await anyio.open_file(path, "wb") as f:
        async for chunk in r.aiter_bytes():
            await f.write(chunk)



import asyncio
import os
from pathlib import Path
from dotenv import load_dotenv

from oe_databank.enums import FileFormat, Frequency, ListingType, Order, SelectionSortOrder, SelectionType, Sequence
from oe_databank.models import DownloadRequestRegion, DownloadRequestVariable, FileDownloadRequestDto, Selection

load_dotenv(Path.home() / ".local/oe.env")

from oe_databank import DatabankAsyncClient



async def main():
    client = DatabankAsyncClient(api_key=os.environ["OE_DATA_API_KEY"])
    print("Fetching data...")
    # await asyncio.gather(
    #     client.list_databanks(to_path="databanks.json"),
    #     client.get_tree("Indicators_GCT", to_path="tree.json"),
    #     client.get_regions("GCT", to_path="regions.json"),
    # )

    await client.query(
        request=FileDownloadRequestDto(
            selections=[
                Selection(
                    selectionType=SelectionType.QUERY,
                    isTemporarySelection=True,
                    databankCode="GCT",
                    sequence=Sequence.EARLIEST_TO_LATEST,
                    groupingMode=False,
                    transposeColumns=False,
                    order=Order.LOCATION_INDICATOR,
                    indicatorSortOrder=SelectionSortOrder.ALPHABETICAL,
                    locationSortOrder=SelectionSortOrder.ALPHABETICAL,
                    format=0,
                    legacyDatafeedFileStructure=False,
                    variables=[
                        DownloadRequestVariable(
                            variableCode="MRSA!$",
                            productTypeCode="GCT",
                            measureCodes=[],
                        )
                    ],
                    regions=[
                        DownloadRequestRegion(
                            databankCode="GCT",
                            regionCode="ETH_ADD",
                        ),
                    ],
                    listingType=ListingType.SHARED,
                    isDataFeed=False,
                    startYear=2000,
                    endYear=2001,
                    precision=5,
                    frequency=Frequency.BOTH,
                    stackedQuarters=False,
                )
            ],
            format=FileFormat.CSV,
        ),
        to_path="data.csv",
    )


asyncio.run(main())



import pytest

from oe_databank.enums import (
    FileFormat,
    Frequency,
    ListingType,
    Order,
    SelectionSortOrder,
    SelectionType,
    Sequence,
)
from oe_databank.models import (
    DownloadRequestRegion,
    DownloadRequestVariable,
    FileDownloadRequestDto,
    QueueDownloadResponse,
    RegionMetadata,
    Selection,
)


class TestFileDownloadRequestDto:

    def test_init_whole_databank(self):
        FileDownloadRequestDto(
            selections=[
                Selection(
                    selectionType=SelectionType.QUERY,
                    isTemporarySelection=True,
                    databankCode="GCT",
                    sequence=Sequence.EARLIEST_TO_LATEST,
                    groupingMode=False,
                    transposeColumns=False,
                    order=Order.LOCATION_INDICATOR,
                    indicatorSortOrder=SelectionSortOrder.ALPHABETICAL,
                    locationSortOrder=SelectionSortOrder.ALPHABETICAL,
                    format=0,
                    legacyDatafeedFileStructure=False,
                    variables=[],
                    regions=[],
                    listingType=ListingType.SHARED,
                    isDataFeed=False,
                    startYear=2000,
                    endYear=2001,
                    precision=5,
                    frequency=Frequency.ANNUAL,
                    stackedQuarters=False,
                )
            ],
            format=FileFormat.CSV,
        )

    def test_init_partial_databank(self):
        FileDownloadRequestDto(
            selections=[
                Selection(
                    selectionType=SelectionType.QUERY,
                    isTemporarySelection=True,
                    databankCode="GCT",
                    sequence=Sequence.EARLIEST_TO_LATEST,
                    groupingMode=False,
                    transposeColumns=False,
                    order=Order.LOCATION_INDICATOR,
                    indicatorSortOrder=SelectionSortOrder.ALPHABETICAL,
                    locationSortOrder=SelectionSortOrder.ALPHABETICAL,
                    format=0,
                    legacyDatafeedFileStructure=False,
                    variables=[
                        DownloadRequestVariable(
                            variableCode="MRSA!$",
                            productTypeCode="GCT",
                            measureCodes=[],
                        )
                    ],
                    regions=[
                        DownloadRequestRegion(
                            databankCode="GCT",
                            regionCode="ETH_ADD",
                        ),
                    ],
                    listingType=ListingType.SHARED,
                    isDataFeed=False,
                    startYear=2000,
                    endYear=2001,
                    precision=5,
                    frequency=Frequency.ANNUAL,
                    stackedQuarters=False,
                )
            ],
            format=FileFormat.CSV,
        )
